package client;

import java.util.*;

import data.*;

public class RobotBrain {

    String id = "";

    private Display  display = null;
    private WorldMap worldMap = null;
    private int initialMoney = 0;
    
    private int col = 0;
    private int row = 0;
    private int score = 0;
    private int money = 0;
    private int capacity = 0;
    private Decision decision = null;
    private GridItem here;
    private Hashtable allPackages = new Hashtable();
    private Hashtable myPackages = new Hashtable();
    private Hashtable theirPackages = new Hashtable();
    private PackageItem delivery = null;
    private boolean headingToBase = false;
    private boolean collectingPackages = false;
    private Path currentPath = null;
    private int bid = 1;

    public RobotBrain(Display d, WorldMap map, String i, int m, int c) {
        display = d;
        worldMap = map;
        id = i;
        money = m;
        initialMoney = m;
        capacity = c;
        display.setMoney(money);
        display.setCapacity(capacity);
    }
    
    public RobotBrain(WorldMap map, String i, int m, int c) {
        worldMap = map;
        id = i;
        money = m;
        initialMoney = m;
        capacity = c;
    }

    public void setXY(int c, int r) {
        col = c;
        row = r;
        worldMap.explored(col, row);
    }

    public int getCol() { return col; }
    public int getRow() { return row; }

    private Random rand = new Random();

    public void pick(String robotID, String pId) {
        GridItem gridItem = worldMap.findRobotGrid(robotID);
        PackageItem pItem = gridItem.pickPackage(pId);
        if (pItem == null) return; // we didn't know about this package
        if (id.equals(robotID)) {
            // It was MINE!
            myPackages.put(pId, pItem);
            if (allPackages.get(pId) != null) {
                allPackages.remove(pId);
                if (display != null) display.setUndeliveredPackages(allPackages.elements());
            }
            if (display != null) display.setCarryPackages(myPackages.elements());
            capacity -= pItem.weight;
            if (display != null) display.setCapacity(capacity);
            if (delivery == null) {
                // lets deliver this package
                delivery = pItem;
                currentPath = new Path(worldMap, col, row, delivery.dCol, delivery.dRow);
                if (display != null) display.setDelivery(delivery.dCol, delivery.dRow);
                if (display != null) display.showStatus("Delivering Package...");
            }
        } else {
            theirPackages.put(pId, pItem);
        }            
    }

    public void sawPackage(PackageItem item) {
        here = worldMap.getGridItem(col, row);
        allPackages.put(item.id, here);
        worldMap.addPackage(item, col, row);
        if (display != null) display.setUndeliveredPackages(allPackages.elements());
    }

    public void drop(String robotID, String pId) {
        Robot robot = Robot.getRobot(robotID);

        if (robot == null) {
            System.out.println("Unknown Robot: " + robotID);
            return;
        }
        if (!robot.me) {
            allPackages.remove(pId);
            return;
        }
        PackageItem pItem = (PackageItem) myPackages.get(pId);
        if (pItem == null) {
            System.out.println("NOT CARRYING " + pId);
            return;
        }
        GridItem gridItem = worldMap.findRobotGrid(robotID);
        gridItem.dropPackage(pItem);
        // System.out.println("DROP " + robotID + " " + pId);
        if (id.equals(robotID)) {
            // It was MINE!
            score += pItem.weight;
            if (display != null) display.setScore(score);
            capacity += pItem.weight;
            if (display != null) display.setCapacity(capacity);
            myPackages.remove(pId);
            if (display != null) display.setCarryPackages(myPackages.elements());
        } else {
            theirPackages.remove(pId);
        }
        if ((delivery != null) && (pItem.id.equals(delivery.id))) {
            // just delivered my package
            delivery = null;
        }

        if ((delivery == null) && (myPackages.size() > 0)) {
            Enumeration e = myPackages.elements();
            delivery = (PackageItem) e.nextElement();
            if (display != null) display.showStatus("Delivering Another Package...");
        }
    }
    
    public void move(String robotID, String dir) {
        Robot robot = Robot.getRobot(robotID);
        if (robot == null) {
            System.out.println("UNKNOWN ROBOT: " + robotID);
            return;
        }

        int oldX = robot.col;
        int oldY = robot.row;
        worldMap.move(robot, dir);
        robot.move(dir);
        
        // if it was about me, update my new position
        if (robot.me) {
            col = robot.col;
            row = robot.row;
            worldMap.explored(col, row);
        }
        
        if (display != null) display.updateGrid(oldX, oldY);
        if (display != null) display.updateGrid(robot.col, robot.row);
    }

    public void addEnemyRobot(String robotID, int rCol, int rRow) {
        if (!Robot.robotExists(robotID)) {
            worldMap.addEnemyRobot(robotID, rCol, rRow);
        }
    }

    public boolean enemyIsNear() {
        boolean near = false;
        for (int c = -1; c < 2; c++) {
            for (int r = -1; r < 2; r++) {
                if (((col + c) < 1) || ((col + c) > worldMap.getWidth()) ||
                    ((row + r) < 1) || ((row + r) > worldMap.getHeight()) ||
                    ((c == 0) && (r == 0))) {
                    continue;
                }
                GridItem item = worldMap.getGridItem(col+c, row+r);
                if (item.hasRobots()) {
                    near = true;
                    break;
                }
            }
        }
        return near;
    }
    
    /*******************************************************
     *
     */

    public String unstick() {
        moveRandomly();
        return decision.toString();
    }

    public String decideWhatToDoNext() {
        decision = null;
        here = worldMap.getGridItem(col, row);

        if (!avoidEnemies() &&
            !exploreBases() &&
            !dropPackages() &&
            !deliverPackages() &&
            !pickPackages() &&
            !collectPackages() && !
            explore()) {
            moveRandomly();
        }

        money -= decision.bid;
        if (display != null) { display.setMoney(money); }

        /*
        try {
            Thread.sleep(500);
        } catch (Exception ignore) {}
        */
        
        return decision.toString();
    }

    private boolean avoidEnemies() {
        if ((currentPath != null) && enemyIsNear() && (bid < 4)) {
            GridItem item = currentPath.lastStep();
            if (item != null) {
                currentPath = new Path(worldMap, col, row, item.col, item.row);
                decision = moveDownPath();
                if (display != null) display.showStatus("Avoiding Enemies...");
            }
        }
        if (decision != null) bid++; else bid = 1;
        
        return (decision != null);
    }
    
    private boolean dropPackages() {
        String pIds = "";
        boolean first = true;
        for (Enumeration p = myPackages.elements(); p.hasMoreElements(); ) {
            PackageItem pItem = (PackageItem) p.nextElement();
            // DROP ALL THE PACKAGES WE CAN DROP 
            if ((col == pItem.dCol) && (row == pItem.dRow)) {
                if (!first) pIds += " ";
                pIds += pItem.id;
                first = false;
            }
        }
        if (!pIds.equals("")) {
            decision = new Decision(bid, Decision.DROP, pIds);
        }
                                    
        return (decision != null);
    }
    
    private boolean pickPackages() {
        Enumeration pElements = here.packageElements();
        
        StringBuffer buf = new StringBuffer();
        boolean one = true;
        int load = 0;
        
        TreeSet set = new TreeSet(new PackageItem());
        for (Enumeration p = pElements; p.hasMoreElements(); ) {
            PackageItem item = (PackageItem) p.nextElement();
            item.cost = (Math.abs(col - item.dCol) + Math.abs(row - item.dRow)) * item.weight;
            set.add(item);
        }

        Iterator iter = set.iterator();
        while (iter.hasNext()) {
            PackageItem item = (PackageItem) iter.next();
            if ((load + item.weight) <= capacity) {
                if (!one) buf.append(" ");
                buf.append(item.id);
                one = false;
                load += item.weight;
            }
        }
        String packageIds = buf.toString();

        if (!packageIds.equals("")) {
            decision = new Decision(bid, Decision.PICK, packageIds);
            if (display != null) display.showStatus("Heading for pickup...");
        }
                
        return (decision != null);
    }

    private GridItem getClosestPackage() {
        int distance = 999999999;
        GridItem best = null;
        for (Enumeration e = allPackages.elements(); e.hasMoreElements(); ) {
            GridItem item = (GridItem) e.nextElement();
            int far = (item.col - col) * (item.col - col) + (item.row - row) * (item.row - row);
            if (far < distance) {
                best = item;
                distance = far;
            }
        }
        return best;
    }
    
    private PackageItem getBestPackage() {
        PackageItem best = null;
        int distance = 999999999;

        for (Enumeration p = myPackages.elements(); p.hasMoreElements(); ) {
            PackageItem pi = (PackageItem) p.nextElement();
            int far = (pi.col - col) * (pi.col - col) + (pi.row - row) * (pi.row - row);
            if ((far < distance) && (pi.weight < capacity)) {
                best = pi;
                distance = far;
            }
        }

        return null;
    }
    
    private boolean exploreBases() {
        if (headingToBase && (currentPath != null)) {
            decision = moveDownPath();
        } else {
            // have we done enough, and are we empty?
            if ((allPackages.size() < 10) && (myPackages.size() == 0)) {
                // look for an empty base we haven't visited
                GridItem newGrid = worldMap.findUnexploredBase();
                if (newGrid != null) {
                    currentPath = new Path(worldMap, col, row, newGrid.col, newGrid.row);
                    if (currentPath.length() > 0) {
                        if (display != null) display.showStatus("Heading to a base...");
                        decision = moveDownPath();
                    }
                }
            }
        }
        headingToBase = (decision != null);
        return headingToBase;
    }

    private boolean deliverPackages() {
        if (delivery == null) {
            // NOT DELIVERING STUFF
            if (myPackages.size() > 0) {
                // BUT ARE CARRYING STUFF
                delivery = getBestPackage();
                currentPath = new Path(worldMap, col, row, delivery.dCol, delivery.dRow);
                if (display != null) display.showStatus("Delivering Another Package...");
                decision = moveDownPath();
            }
        } else {
            if (currentPath == null) {
                currentPath = new Path(worldMap, col, row, delivery.dCol, delivery.dRow);
                decision = moveDownPath();
            } else {
                // HAVE SOMETHING, KEEP GOING
                decision = moveDownPath();
            }
        }
        return (decision != null);
    }

    private Decision moveDownPath() {

        if (currentPath == null) return null;
        
        GridItem next = currentPath.nextStep();
        if (currentPath.length() == 0) {
            if (headingToBase) System.out.println("REACHED BASE");
            currentPath = null;
        }

        if (next == null) {
            return null;
        }
        
        boolean destIsEast  = (col < next.col);
        boolean destIsWest  = (col > next.col);
        boolean destIsNorth = (row < next.row);
        boolean destIsSouth = (row > next.row);

        if (destIsNorth) decision = new Decision(bid, Decision.MOVE, "N");
        if (destIsSouth) decision = new Decision(bid, Decision.MOVE, "S");
        if (destIsEast) decision = new Decision(bid, Decision.MOVE, "E");
        if (destIsWest) decision = new Decision(bid, Decision.MOVE, "W");

        return decision;
    }

    private boolean collectPackages() {
        if (collectingPackages && (currentPath != null)) {
            decision = moveDownPath();
        } else {
            // DID WE LEAVE SOME PACKAGES BEHIND?
            if (allPackages.size() > 0) {
                GridItem goods = getClosestPackage();
                if (goods != null) {
                    currentPath = new Path(worldMap, col, row, goods.col, goods.row);
                    if (currentPath.length() > 0) {
                        decision = moveDownPath();
                    } else {
                        currentPath = null;
                    }
                }
            }
        }
        collectingPackages = (decision != null);
        return collectingPackages;
    }
        
    private boolean explore() {
        if (currentPath != null) {
            decision = moveDownPath();
        } else {
            GridItem newGrid = worldMap.getUnexploredGridItem(col, row);
            if (newGrid != null) {
                //            System.out.println("EXPLORE TOWARDS " + newGrid);
                if (display != null) display.exploreTowards(newGrid.col, newGrid.row);
                newGrid.explored = true; // just in case we can't get there
                currentPath = new Path(worldMap, col, row, newGrid.col, newGrid.row);
                if (currentPath.length() > 0) {
                    decision = moveDownPath();
                } else {
                    currentPath = null;
                }
            }
        }
        return (decision != null);
    }
    
    private boolean moveRandomly() {
        boolean directionOK = false;
        GridItem there = null;

        String[] dirs = new String[4];
        dirs[0] = "N";
        dirs[1] = "S";
        dirs[2] = "W";
        dirs[3] = "E";
        
        while (!directionOK) {
            int r = rand.nextInt(4);
            decision = new Decision(bid, Decision.MOVE, dirs[r]);
            directionOK = okDirection(decision);
        }

        if (display != null) display.showStatus("Moving Randomly...");
        
        return true;
    }

    private boolean okDirection(Decision decision) {
        int r = 0;
        int c = 0;
        
        if (decision.args.equals("N")) r =  1;
        if (decision.args.equals("S")) r = -1;
        if (decision.args.equals("E")) c =  1;
        if (decision.args.equals("W")) c = -1;

        if (((row+r) < 1) || ((row+r) > worldMap.getHeight()) ||
            ((col+c) < 1) || ((col+c) > worldMap.getWidth())) {
            // don't ever go off the map
            return false;
        }
        
        boolean directionOK = true;
        GridItem gridItem = worldMap.getGridItem(col+c, row+r);
            
        directionOK = !gridItem.isImpassible();
        
        return directionOK;
    }
}
