package client;

import java.util.*;
import java.io.*;
import java.net.*;

import javax.swing.*;
import javax.swing.border.*;

import data.Robot; // wow, there is a Robot class in java.awt!!
import java.awt.*;

import data.*;

public class RobotController implements Display {

    protected static int serverPort = 9000;
    protected static String serverHost = "localhost";

    protected static PrintWriter writer = null;

    private static final int STATE_LOADING_MAP       = 1;
    private static final int STATE_LOADING_CONFIG    = 2;
    private static final int STATE_GETTING_LOCATIONS = 3;
    private static final int STATE_TURN_START        = 4;
    private static final int STATE_SENDING_COMMAND   = 5;
    private static final int STATE_SERVER_RESPONSE   = 6;

    private int state = STATE_LOADING_MAP;

    private static boolean debug = false;

    private WorldMap worldMap = null;
    private RobotBrain brain = null;
    private String myRobotID = "";
    private boolean lastAction = false;
    private static boolean showUI = false;

    private String lastServerMessage = "";
    private String lastClientMessage = "";

    private JFrame       frame;
    private JPanel       worldMapPanel;
    private JLabel       scoreLabel;
    private JLabel       moneyLabel;
    private JLabel       capacityLabel;
    private JList        packageCarryList;
    private JList        packageUndeliveredList;
    private JTextField   robotStatusField;
    private JProgressBar pathLengthBar;
    private JLabel[][]   mapCells;
    private Color        background;
    
    public static void main(String[] args) {
        for (int i = 0; i < args.length; i++) {
            if (args[i].equalsIgnoreCase("-p")) {
                serverPort = new Integer(args[++i]).intValue();
            } else
            if (args[i].equalsIgnoreCase("-ui")) {
                showUI = true;
            } else
            if (args[i].equalsIgnoreCase("-h")) {
                serverHost = args[++i];
            } else
            if (args[i].equalsIgnoreCase("-d")) {
                debug = true;
            }
        }

        try {
            new RobotController();
        } catch (Exception bad) {
            bad.printStackTrace();
        }
    }

    public RobotController() throws Exception {
        if (showUI) {
            buildUI();
        }
        new ReaderThread().start();  // start the socket listener
    }

    private String [] SERVER_MESSAGE = {"died",
                                        "Game over",
                                        "***" };

    private void onServerMessage(String message) throws Exception {
        if (debug) {
            System.out.println("server said [" + message + "]");
        }

        for (int m = 0; m < SERVER_MESSAGE.length; m++) {
            if (message.indexOf(SERVER_MESSAGE[m]) != -1) {
                System.out.println(message);
                return;
            }
        }

        switch (state) {
        case STATE_LOADING_MAP:
            loadMap(message);
            break;

        case STATE_LOADING_CONFIG:
            loadConfig(message);
            break;
            
        case STATE_GETTING_LOCATIONS:
            gettingLocations(message);
            break;
            
        case STATE_TURN_START:
            getPackages(message);
            break;

        case STATE_SENDING_COMMAND:
            break;
            
        case STATE_SERVER_RESPONSE:
            processServerResponse(message);
            break;
        }
    }
    
    private void onServerError(String message) {
        showStatus("ERROR " + message);
    }

    public class ReaderThread extends Thread {
        public ReaderThread() {
            super();
        }
        
        public void run() {
            try {
                System.out.println("Connect " + serverHost + " " + serverPort);
                Socket socket = new Socket(serverHost, serverPort);

                System.out.println("Client Initializing....");
                writer = new PrintWriter(socket.getOutputStream());

                writer.println("Player");
                writer.flush();
            
                BufferedReader reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));

                System.out.println("Waiting for Server...");
                lastServerMessage = reader.readLine();
                while (true) {
                    if (lastServerMessage == null) break;
                    onServerMessage(lastServerMessage);
                    lastServerMessage = reader.readLine();
                    if (lastServerMessage == null) break;
                }

            } catch (Exception err) {
                System.out.println("MY LAST MESSAGE: " + lastClientMessage);
                System.out.println("SERVER SAID: " + lastServerMessage);
                err.printStackTrace();
                onServerError(lastServerMessage + " " + err.getMessage());
            }
            System.exit(0);
        }
    }

    // -------------------------------------------------------
    int width = -1;
    int height = -1;
    int row = 0;
    private void loadMap(String line) throws Exception {
        if (width == -1) {
            StringTokenizer st = new StringTokenizer(line, " ");
            width = new Integer(st.nextToken()).intValue();
            height = new Integer(st.nextToken()).intValue();
            row = 1;
        } else {
            if (row < 0) {
                System.out.println("UNKNOWN [" + line + "]");
            } else {
                if (worldMap == null) {
                    worldMap = new WorldMap(width, height);
                }
                for (int col = 1; col <= line.length(); col++) {
                    worldMap.addItem(col, row, line.charAt(col-1));
                }
            
                row++;
            }
        }

        if (row > height) {
            if (showUI) createWorldMap();
            state = STATE_LOADING_CONFIG;
        }
    }

    private void loadConfig(String line) {
        StringTokenizer st = new StringTokenizer(line, " ");

        String id = st.nextToken();
        int capacity = new Integer(st.nextToken()).intValue();
        int money = new Integer(st.nextToken()).intValue();

        if (showUI) {
            brain = new RobotBrain(this, worldMap, id, money, capacity);
        } else {
            brain = new RobotBrain(worldMap, id, money, capacity);
        }

        //System.out.println("LOADED " + id + " Capacity = " + capacity +
        //                           "  Money = " + money);
        showStatus("Loaded our config");
        state = STATE_GETTING_LOCATIONS;
    }

    private void gettingLocations(String line) {
        StringTokenizer st = new StringTokenizer(line, " ");

        boolean first = true;
        while (st.hasMoreTokens()) {
            String id = st.nextToken();
            if (id.startsWith("#")) {
                id = id.substring(1);
            }
            String xDelim = st.nextToken();
            int col = new Integer(st.nextToken()).intValue();
            String yDelim = st.nextToken();
            int row = new Integer(st.nextToken()).intValue();

            brain.setXY(col, row);
            if (first) {
                Robot robot = worldMap.addRobot(id, col, row);
                System.out.println("MY ROBOT " + id + " @ " + col + "," + row);
                myRobotID = id;
                first = false;
            } else {
                worldMap.addEnemyRobot(id, col, row);
                System.out.println("ENEMY ROBOT " + id + " @ " + col + "," + row);
            }
        }
        
        showStatus("Got Initial Locations");
        state = STATE_TURN_START;
    }
    
    private void getPackages(String line) {
        StringTokenizer st = new StringTokenizer(line, " ");

        while (st.hasMoreTokens()) {
            String packID = st.nextToken();
            int col = new Integer(st.nextToken()).intValue();
            int row = new Integer(st.nextToken()).intValue();
            int w = new Integer(st.nextToken()).intValue();
            brain.sawPackage(new PackageItem(packID, w,
                                             brain.getCol(),
                                             brain.getRow(), col, row));

        }

        //        showStatus("Got Package Message: " + line);
        state = STATE_SENDING_COMMAND;

        // DO THE NEXT THING
        //System.out.println("Decide what to do...");
        if (!lastAction) {
            lastClientMessage = brain.unstick();
        } else {
            lastClientMessage = brain.decideWhatToDoNext();
        }
        //System.out.println("SAY:" + lastClientMessage);
        writer.println(lastClientMessage);
        writer.flush();

        state = STATE_SERVER_RESPONSE;
    }
    
    private void processServerResponse(String line) {
        //System.out.println("server said:" + line);
        StringTokenizer st = new StringTokenizer(line, " ");

        boolean me = false;
        lastAction = false;
        String robotID = "";
        while (st.hasMoreTokens()) {
            String token = st.nextToken();
            if (token.startsWith("#")) {
                robotID = token.substring(1);
                if (robotID.equals(myRobotID)) {
                    me = true;
                }
            } else
            if (token.equals("N") || token.equals("S") ||
                token.equals("E") || token.equals("W")) {
                brain.move(robotID, token);
                if (me) lastAction = true;
            } else
            if (token.equals("P")) {
                brain.pick(robotID, st.nextToken());
                if (me) lastAction = true;
            } else
            if (token.equals("D")) {
                brain.drop(robotID, st.nextToken());
                if (me) lastAction = true;
            } else
            if (token.equals("X")) {
                int c = new Integer(st.nextToken()).intValue();
                st.nextToken();
                int r = new Integer(st.nextToken()).intValue();
                brain.addEnemyRobot(robotID, c, r);
            }
        }

        //        System.out.println("GOT RESPONSE\n" + worldMap);
        state = STATE_TURN_START;
    }
    
    private void buildUI() {
        final String systemLookAndFeel = UIManager.getSystemLookAndFeelClassName();
        final String genericLookAndFeel = UIManager.getCrossPlatformLookAndFeelClassName();
        try {
            String className;
            if (systemLookAndFeel != null) {
                className = systemLookAndFeel;
            } else {
                className = genericLookAndFeel;
            }
            UIManager.setLookAndFeel(className);
        } catch(Exception ex) {
            ex.printStackTrace();
        }

        frame = new JFrame("Robot World");
        JPanel mainPanel = new JPanel(new BorderLayout(5,5));
        mainPanel.setBorder(new EmptyBorder(5,5,5,5));
        frame.getContentPane().setLayout(new BorderLayout(0,0));
        frame.getContentPane().add(mainPanel, "Center");

        JPanel bottomPanel = new JPanel(new BorderLayout(5,5));
        JPanel rightPanel = new JPanel(new BorderLayout(5,5));
        JPanel rightTopPanel = new JPanel(new GridLayout(5,1));
        JPanel scorePanel = new JPanel(new GridLayout(1,2));
        JPanel moneyPanel = new JPanel(new GridLayout(1,2));
        JPanel capacityPanel = new JPanel(new GridLayout(1,2));
        JPanel packagePanel = new JPanel(new BorderLayout(5,5));
        JPanel packageCarryPanel = new JPanel(new BorderLayout(5,5));
        JPanel packageUndeliveredPanel = new JPanel(new BorderLayout(5,5));
        
        worldMapPanel = new JPanel();
        background = worldMapPanel.getBackground();

        scoreLabel = new JLabel("0", SwingConstants.RIGHT);
        moneyLabel = new JLabel("0", SwingConstants.RIGHT);
        capacityLabel = new JLabel("0", SwingConstants.RIGHT);

        JLabel scoreTitle = new JLabel("Score");
        JLabel moneyTitle = new JLabel("Money");
        JLabel capacityTitle = new JLabel("Capacity");

        scorePanel.add(scoreTitle);
        scorePanel.add(scoreLabel);
        
        moneyPanel.add(moneyTitle);
        moneyPanel.add(moneyLabel);

        capacityPanel.add(capacityTitle);
        capacityPanel.add(capacityLabel);
        
        rightTopPanel.add(scorePanel);
        rightTopPanel.add(moneyPanel);
        rightTopPanel.add(capacityPanel);

        JLabel packageCarryListTitle = new JLabel("Carrying Packages");
        packageCarryList = new JList(new Vector());
        packageCarryList.setMinimumSize(new Dimension(200, 400));
        packageCarryList.setFixedCellWidth(200);
        packageCarryList.setVisibleRowCount(5);
        JScrollPane packageCarryScroll = new JScrollPane(packageCarryList, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        packageCarryPanel.add(packageCarryListTitle, "North");
        packageCarryPanel.add(packageCarryScroll, "Center");
        
        JLabel packageUndeliveredListTitle = new JLabel("Undelivered Packages");
        packageUndeliveredList = new JList(new Vector());
        packageUndeliveredList.setPreferredSize(new Dimension(200, 400));
        packageUndeliveredList.setFixedCellWidth(200);
        packageUndeliveredList.setVisibleRowCount(8);
        JScrollPane packageUndeliveredScroll = new JScrollPane(packageUndeliveredList, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        packageUndeliveredPanel.add(packageUndeliveredListTitle, "North");
        packageUndeliveredPanel.add(packageUndeliveredScroll, "Center");

        packagePanel.add(packageCarryPanel, "Center");
        packagePanel.add(packageUndeliveredPanel, "South");
        
        rightPanel.add(rightTopPanel, "North");
        rightPanel.add(packagePanel, "Center");

        robotStatusField = new JTextField();
        pathLengthBar = new JProgressBar(0, 100);
        bottomPanel.add(robotStatusField, "Center");
        bottomPanel.add(pathLengthBar, "East");

        mainPanel.add(worldMapPanel, "Center");
        mainPanel.add(rightPanel, "East");
        mainPanel.add(bottomPanel, "South");

        frame.setLocation(50,50);
        frame.setSize(600,500);

        Path.display = this;
        
        frame.show();
    }

    public void updateGrid(int col, int row) {
        Color color = worldMap.getGridItem(col, row).getColor();
        //        System.out.println(" " + (color == Color.red) + " " + col + "," + row);
        mapCells[col-1][height - row].setBackground(color);
    }
    
    public void setDelivery(int col, int row) {
        mapCells[col-1][height - row].setBackground(Color.pink);
    }

    public void setFrontier(int col, int row) {
        mapCells[col-1][height - row].setBackground(Color.cyan);
    }
    public void setPath(int col, int row) {
        mapCells[col-1][height - row].setBackground(Color.white);
    }

    public void exploreTowards(int col, int row) {
        showStatus("Exporing...");
        mapCells[col-1][height - row].setBackground(Color.white);
    }

    public void setMoney(int money) {
        moneyLabel.setText(new Integer(money).toString());
    }

    public void setCapacity(int capacity) {
        capacityLabel.setText(new Integer(capacity).toString());
    }

    public void setScore(int score) {
        scoreLabel.setText(new Integer(score).toString());
    }

    public void setPathLength(int len) {
        if (len > pathLengthBar.getMaximum()) {
            pathLengthBar.setMaximum(7*len/6);
        }
        pathLengthBar.setValue(len);
    }

    public void setCarryPackages(Enumeration pElements) {
        Vector packages = new Vector();
        for (Enumeration p = pElements; p.hasMoreElements(); ) {
            PackageItem item = (PackageItem) p.nextElement();
            packages.addElement(item.toString());
        }
        packageCarryList.setListData(packages);
    }
    
    public void setUndeliveredPackages(Enumeration pElements) {
        Vector packages = new Vector();
        for (Enumeration g = pElements; g.hasMoreElements(); ) {
            GridItem gi = (GridItem) g.nextElement();
            for (Enumeration p = gi.packageElements(); p.hasMoreElements(); ) {
                PackageItem item = (PackageItem) p.nextElement();
                packages.addElement(item.toString());
            }
        }
        packageUndeliveredList.setListData(packages);
    }
    
    public void showStatus(String message) {
        if (robotStatusField != null) {
            robotStatusField.setText(message);
        } else {
            System.out.println(message);
        }
    }
    
    public void createWorldMap() {
        worldMapPanel.setLayout(new GridLayout(height, width, 1,1));

        int cellW = 500/width;
        int cellH = 250/height;

        mapCells = new JLabel[width][height];
        
        for (int row = worldMap.getHeight(); row >= 1; row--) {
            for (int col = 1; col <= worldMap.getWidth(); col++) {
                GridItem item = worldMap.getGridItem(col, row);
                JLabel label = new JLabel(" ");
                label.setOpaque(true);
                label.setPreferredSize(new Dimension(cellW, cellH));
                label.setMinimumSize(new Dimension(cellW, cellH));
                label.setMaximumSize(new Dimension(cellW, cellH));
                label.setBackground(item.getColor());
                worldMapPanel.add(label);
                mapCells[col-1][height - row] = label;
            }
        }
    }
}
