package server;

import java.io.*;
import java.net.*;

public class RobotClient {

    private String robotID;
    private WorldServer server = null;
    private Socket socket = null;
    
    protected PrintWriter writer = null;
    
    public RobotClient(WorldServer srv, String id, Socket sock) {
        robotID = id;
        server = srv;
        socket = sock;
        new ReaderThread().start();
    }

    public String getID() { return robotID; }
    
    public void sendMessage(String message) {
        writer.println(message);
        writer.flush();
    }
    
    public class ReaderThread extends Thread {
        public void run() {
            try {
                BufferedReader reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
                writer = new PrintWriter(socket.getOutputStream());
                
                System.out.println("Wait for input from " + robotID + "...");
                String line = reader.readLine();
                while (true) {
                    server.onClientMessage(robotID, line);
                    line = reader.readLine();
                    if (line == null) break;
                }

                System.out.println("Client done " + robotID);
            } catch (Exception err) {
                server.onClientError(robotID, err.getMessage());
            }
        }
    }
}
